/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM                                   |
 *| (C) Copyright IBM Corp. 2004.  All Rights Reserved.                    |
 *|                                                                        |
 *| US Government Users Restricted Rights - Use, duplication or disclosure |
 *| restricted by GSA ADP Schedule Contract with IBM Corp.                 |
 *+------------------------------------------------------------------------+
 */
package com.ibm.jps.uctoanalysis.transformation.rules;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gmf.runtime.diagram.ui.services.layout.LayoutType;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.uml2.uml.Abstraction;
import org.eclipse.uml2.uml.Actor;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Profile;
import org.eclipse.uml2.uml.Stereotype;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.UseCase;

import com.ibm.jps.util.RSMDiagram;
import com.ibm.jps.util.RSMModel;
import com.ibm.jps.util.RSMResource;
import com.ibm.jps.util.RSMUtil;
import com.ibm.xtools.modeler.ui.UMLModeler;
import com.ibm.xtools.transform.core.ITransformContext;
import com.ibm.xtools.uml.transform.core.IsElementKindCondition;
import com.ibm.xtools.uml.ui.diagram.IUMLDiagramHelper;
import com.ibm.xtools.umlnotation.UMLDiagramKind;

/*
 * (non-Javadoc)
 * 
 * @see com.ibm.xtools.transform.core.AbstractRule
 */
public class ModelRule extends com.ibm.xtools.transform.core.ModelRule {
	Model inputModel = null;
	Model outputModel = null;
	Package ucReaPkg = null;
	Package designLayer = null;
	List<Diagram> dgms = null;
	
	/**
	 * Default constructor to create a rule.
	 */
	public ModelRule() {
		super();
        setAcceptCondition(new IsElementKindCondition(UMLPackage.eINSTANCE.getModel()));
	}

	/**
	 * constructor
	 * @param id
	 * @param name
	 */
	public ModelRule(String id, String name) {
		super(id, name);
        setAcceptCondition(new IsElementKindCondition(UMLPackage.eINSTANCE.getModel()));
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see com.ibm.xtools.transform.core.AbstractRule#createTarget(com.ibm.xtools.transform.core.ITransformContext)
	 */
	public Object createTarget(ITransformContext ruleContext) throws CoreException {
        inputModel = (Model) ruleContext.getSource();
        outputModel = (Model) ruleContext.getTargetContainer();
        log("--- Processing model " + inputModel.getName() + " > " + outputModel.getName());
        
        // Apply the RUP Analysis profile
        if (outputModel.getAppliedProfile("Analysis Profile") == null) {
        	Profile analysisProfile = RSMModel.findProfile("com.ibm.xtools.modeler.ui.templates", "Analysis Profile");
        	if (analysisProfile != null) {
        		outputModel.applyProfile(analysisProfile);
        	}
        }
        
        // Apply the ArchitectAssistant profile
        if (outputModel.getAppliedProfile("ArchitectAssistant") == null) {
        	Profile archProfile = RSMModel.findProfile("com.ibm.jps.architect", "ArchitectAssistant");
        	if (archProfile != null) {
        		outputModel.applyProfile(archProfile);
        	}
        }
        
        // Process all use cases in the input model
        List<EObject> allUCs = RSMModel.findElements(inputModel, UMLPackage.eINSTANCE.getUseCase());
        dgms = new ArrayList<Diagram>();
        for (Object curUC_o:allUCs) {
		    processUC((UseCase) curUC_o);
		}
        
        
        // Because SWT will trigger an SWTException for any calls made from a non-UI thread that 
        // must be made from the UI thread, need to update the diagram using the code below. See
        // also Eclipse Bug 57384.
		final IWorkbench workbench = PlatformUI.getWorkbench();
		workbench.getDisplay().asyncExec(new Runnable() {
			public void run() {
				IWorkbenchWindow window = workbench.getActiveWorkbenchWindow();
				if (window != null) {
					updateDiagramLayout();
				}
			}
		});  
		
		return null;
	}

	private synchronized void updateDiagramLayout() {
		IUMLDiagramHelper helper = UMLModeler.getUMLDiagramHelper();
        for (Diagram curDgm:dgms) {
        	helper.layout(curDgm, LayoutType.DEFAULT);
		}
	}
	
	private void processUC(UseCase inUC) {
		Stereotype st = null;
		String ucId = RSMResource.getId(inUC);
     	log("    Processing use case " + inUC.getName());
     	
     	// UC Realization package
     	if (ucReaPkg == null) {
     		String ucReaId = "UCR " + RSMResource.getId(inputModel);
        	ucReaPkg = (Package) findElement(outputModel, UMLPackage.eINSTANCE.getPackage(), ucReaId);
        	if (ucReaPkg == null) {
        		ucReaPkg = (Package) outputModel.createPackagedElement("UC Realizations", UMLPackage.eINSTANCE.getPackage());
        		RSMModel.setDocumentation(ucReaPkg, "@generated - do not write past this point " + ucReaId + " (UC Realizations)");
        	}
     	}
     	// Design layer
     	if (designLayer == null) {
     		String designLayerId = "UCP " + RSMResource.getId(inputModel);
        	designLayer = (Package) findElement(outputModel, UMLPackage.eINSTANCE.getPackage(), designLayerId);
        	if (designLayer == null) {
        		designLayer = (Package) outputModel.createPackagedElement("Design", UMLPackage.eINSTANCE.getPackage());
        		RSMModel.setDocumentation(designLayer, "@generated - do not write past this point " + designLayerId + " (Design)");
        	}
     	}
     	
     	// Create a UC Package in the Use-Case Realization package
    	String name = RSMUtil.convertToCamelCase(inUC.getName());
    	String pkgId = "UCP " + ucId;
    	Package pkg = (Package) findElement(outputModel, UMLPackage.eINSTANCE.getPackage(), pkgId);
    	if (pkg == null) {
    		pkg = (Package) ucReaPkg.createPackagedElement(inUC.getName(), UMLPackage.eINSTANCE.getPackage());
    		RSMModel.setDocumentation(pkg, "@generated - do not write past this point " + pkgId + " (" + inUC.getName() + ")");
    	} else {
    		handleNameChange(pkg, inUC.getName());
    	}
    	
    	// Create a Main diagram in the UC Package if it does not exist
    	IUMLDiagramHelper helper = UMLModeler.getUMLDiagramHelper();
    	Diagram diag = RSMDiagram.findDiagram(pkg, "Main", UMLDiagramKind.CLASS_LITERAL);
    	if (diag == null) {
    		diag = helper.createDiagram(pkg, UMLDiagramKind.CLASS_LITERAL);
    		diag.setName("Main");
    		helper.setMainDiagram(pkg, diag);
    	}

    	// Find UC node
    	Node ucNode = RSMDiagram.findNode(diag, inUC);
    	if (ucNode == null) {
    		ucNode = helper.createNode(diag, inUC);
    	}
    	
		// Create a controller class if it does not already exist.
    	String controllerId = "UCC " + ucId;
    	String controllerName = name + "Controller";
    	Class cl = (Class) findElement(outputModel, UMLPackage.eINSTANCE.getClass_(), controllerId);
    	if (cl == null) {
    		cl = designLayer.createOwnedClass(controllerName, false);
    		st = cl.getApplicableStereotype("RUPAnalysis::Control");
    		if (st != null) cl.applyStereotype(st);
    		RSMModel.setDocumentation(cl, "@generated - do not write past this point " + controllerId + " (" + controllerName + ")");
    	} else {
    		handleNameChange(cl, controllerName);
    	}
    	
    	// Create a <<trace>> relationship to the UC
    	Abstraction trace = RSMModel.findTrace(cl, inUC);
    	if (trace == null) {
    		trace = RSMModel.createTrace(cl, inUC);
    	}
    	
    	// ... and add it to the main diagram
    	Node node = RSMDiagram.findNode(diag, cl);
    	if (node == null) {
    		node = helper.createNode(diag, cl);
    		helper.createEdge(node, ucNode, trace);
    	}
    	
    	// For each Actor-UC association, create a boundary class
    	List<Actor> actors = RSMModel.findAssociatedActors(inputModel, inUC);
    	for (Actor curActor : actors) {
    		String boundaryId = "UCB " + RSMResource.getId(curActor);
    		String boundaryName = RSMUtil.convertToCamelCase(curActor.getName());
    		if (!curActor.getKeywords().contains(new String("system"))) {
    			boundaryName += name + "Form";
        		boundaryId += " " + ucId;
    		}
     		cl = (Class) findElement(outputModel, UMLPackage.eINSTANCE.getClass_(), boundaryId);
    		if (cl == null) {
    			cl = designLayer.createOwnedClass(boundaryName, false);
        		st = cl.getApplicableStereotype("RUPAnalysis::Boundary");
        		if (st != null) cl.applyStereotype(st);
        		RSMModel.setDocumentation(cl, "@generated - do not write past this point " + boundaryId + " (" + boundaryName + ")");
    	   	} else {
        		handleNameChange(cl, boundaryName);
    		}
    		trace = RSMModel.findTrace(cl, inUC);
        	if (trace == null) {
        		trace = RSMModel.createTrace(cl, inUC);
        	}
        	node = RSMDiagram.findNode(diag, cl);
        	if (node == null) {
        		node = helper.createNode(diag, cl);
        		helper.createEdge(node, ucNode, trace);
        	}
    	} 	
    	dgms.add(diag);
	}
	
    public Element findElement(Package inPkg, EClass inElementKind, String inElementId) {
    	List<EObject> temp = RSMModel.findElements(inPkg, inElementKind);
		for (EObject curObj : temp) {
			if (curObj.eClass() == inElementKind) {
				Element curElt = (Element) curObj;
				String doc = RSMModel.getDocumentation(curElt);
				if (!doc.equals("") && doc.contains(inElementId)) {
					return curElt;
				}
			}
		}
		return null;
	}  
    
    /**
     * handleNameChange detects whether or not there was a change of name in the input element (e.g. a change
     * of the name of the actor or use case). The documentation of the matching element in the output model (inNE)
     * contains the original name. Therefore if this original name is different from the new target name, this 
     * means that the input element has been renamed. The documentation of the output element is updated 
     * accordingly. However the actual name of the output element will be NOT changed IF different from
     * the target name, as it indicates that the user has manually changed the name of the output element.
     * 
     * @param inNE - the current named element in the output model
     * @param targetName - the name that inNE should have except if changed by the user
     */
    private void handleNameChange(NamedElement inNE, String targetName) {
    	String doc = RSMModel.getDocumentation(inNE);
 		if (doc != null && !doc.equals("")) {
 		   	int posStart = doc.lastIndexOf("(");
 	    	int posEnd = -1;
			if (posStart > 0) {
				posEnd = doc.lastIndexOf(")");
				if (posEnd > posStart+1) {
					String originalName = doc.substring(posStart+1, posEnd);
					if (!targetName.equals(originalName)) {
						if (inNE.getName().equals(originalName)) inNE.setName(targetName);
			    		RSMModel.setDocumentation(inNE, doc.substring(0,posStart+1)+targetName+")");
			    	}
				}
			}
		}
    }
    
	private void log(String inMsg) {
		System.out.println(inMsg);
	}

}
